#include <vector>
#include <string>
#include <fstream>
#include <cmath>
#include <iostream>
#include <sstream>
#include <cstdlib>
#include <algorithm>
#include <cstring>
#include <limits>

#include <CL/opencl.h>

#include "NN.h"
#include "util.h"
#include "PSO.h"

using namespace std;

#define PRINT_INFORMATION 1
#define PRINT_DETAIL_TIMING_INFORMATION 0
#define PRINT_GAP 100

#define CL_FINISH clFinish(queue_);
//#define CL_FINISH

float sigmoid(float x)
{
    return 1 / (1 + expf(-x));
}

// NN

/** Reads in network layer configuration from file to host vector h_layers_ */
void NN::readInNNetConfig(string fileName)
{
    ifstream file(fileName.c_str());
    if(!file.is_open()) error(string("Error opening file: ") + fileName, -3);

    //read layers from file to host vector
    int layer;
    while(file >> layer)
        h_layers_.push_back(layer);
}

/** Reads in training data from file to host vectors h_*_inputs_ and h_*_targets_ */
void NN::readInTrainingData(string fileName)
{
    ifstream file(fileName.c_str());
    if(!file.is_open()) error(string("Error opening file: ") + fileName, -7);

    // percetage of the samples which is gonna be used as validation ( [0,100> )
    // 0 means no validation
    float trainingRatio, validationRatio, testRatio;
    file >> trainingRatio >> validationRatio >> testRatio;

    // number of inputs and outputs
    int num_in, num_out;
    file >> num_in >> num_out;

    if(num_in != getInputSize()) error("Input size from config file does not match data input size", -149);
    if(num_out != getOutputSize()) error("Output size from config file does not match data output size", -149);

    // input and ouput ranges
    in_ranges_ = vector<float>(2*num_in);
    for(int i=0;i<num_in;i++)
        file >> in_ranges_[2*i] >> in_ranges_[2*i+1];

    h_out_ranges_ = vector<float>(2 * num_out);
    for(int i=0;i<num_out;i++)
        file >> h_out_ranges_[2*i] >> h_out_ranges_[2*i+1];

    // training/validation/test samples
    vector<float> inputs;
    vector<float> targets;
    float tmp_num;
    int numSamples=0;
    while(file >> tmp_num)
    {
        // read input
        inputs.push_back(tmp_num);
        for(int i=1;i<num_in;i++)
        {
            file >> tmp_num;
            inputs.push_back(tmp_num);
        }

        // read output
        for(int i=0;i<num_out;i++)
        {
            file >> tmp_num;
            targets.push_back(tmp_num);
        }
        numSamples++;
    }
    inputs = scaleInputData(inputs);

    // Samples are shuffled (in case they are sorted)
    //shuffleSamples(inputs, targets, num_in, num_out);

    // divide the inputs into training set, validation set and test set
    int sum = trainingRatio + validationRatio + testRatio;
    int trainingSetSize, validationSetSize, testSetSize;
    trainingSetSize = floor(trainingRatio/sum * numSamples + 0.5); // round
    validationSetSize = floor(validationRatio/sum * numSamples + 0.5);
    testSetSize = numSamples - trainingSetSize - validationSetSize; // what's left

    if(trainingSetSize <= 0 || trainingSetSize > numSamples)
        error("Wrong training/validation/test set size", 173);

    int trainInputCopySize = getInputSize() * trainingSetSize;
    int trainTargetCopySize = getOutputSize() * trainingSetSize;

    int validInputCopySize = getInputSize() * validationSetSize;
    int validTargetCopySize = getOutputSize() * validationSetSize;


    // training set is in the beginning, validation set in the middle, test set in the end
    h_training_inputs_ = vector<float>(inputs.begin(), inputs.begin() + trainInputCopySize);
    h_training_targets_ = vector<float>(targets.begin(), targets.begin() + trainTargetCopySize);

    h_validation_inputs_ = vector<float>(inputs.begin() + trainInputCopySize, inputs.begin() + trainInputCopySize + validInputCopySize);
    h_validation_targets_ = vector<float>(targets.begin() + trainTargetCopySize, targets.begin() + trainTargetCopySize + validTargetCopySize);

    h_test_inputs_ = vector<float>(inputs.begin() + trainInputCopySize + validInputCopySize, inputs.end());
    h_test_targets_ = vector<float>(targets.begin() + trainTargetCopySize + validTargetCopySize, targets.end());

    if(getNumTrainingSamples() + getNumValidationSamples() + getNumTestSamples() != numSamples)
        error("Something is wrong with the samples", 136);
}

void NN::shuffleSamples(vector<float> &inputs, vector<float> &targets, int inputSize, int targetSize)
{
    int numSamples = inputs.size()/inputSize;
    vector<int> indices;
    for(int i=0; i<numSamples;i++) indices.push_back(i);
    random_shuffle(indices.begin(),indices.end());

    for(int i=0;i<numSamples;i++)
    {
        // swap inputs
        for(int j=0; j<inputSize; j++)
            swap(inputs[i*inputSize + j], inputs[indices[i]*inputSize + j]);

        // swap outputs
        for(int j=0; j<targetSize; j++)
            swap(targets[i*targetSize + j], targets[indices[i]*targetSize + j]);
    }

}

/** Scales several input data samples which are in ranges in_ranges_ to [-1,1]*/
vector<float> NN::scaleInputData(vector<float> in)
{
    for(size_t i=0;i<in.size() / getInputSize();i++)
        for(int j=0;j<getInputSize();j++)
        {
            in[i*getInputSize()+j] = 2 * (in[i*getInputSize()+j] - in_ranges_[2*j]) / (in_ranges_[2*j+1] - in_ranges_[2*j]) - 1;
        }
    return in;
}


// Scales output (e.g. targets) to a range [0,1];
vector<float> NN::inverseScaleOutputData(const vector<float> &out)
{
    vector<float> output(out);
    for(int i=0;i<getOutputSize();i++)
        output[i] = (output[i] - h_out_ranges_[2*i]) / (h_out_ranges_[2*i+1] - h_out_ranges_[2*i]);
    return output;
}

NN::NN(string nnetConfigFile, string trainingDataFile)
{
    readInNNetConfig(nnetConfigFile); //h_layers_
    readInTrainingData(trainingDataFile); // h_inputs_ , h_targets_

    // allocate memory for CPU simulation
    h_output_ = new float[getWidestLayer()];
    h_temp_output_ = new float[getWidestLayer()];
}

NN::NN()
{
}
NN::~NN()
{
    delete[] h_output_;
    delete[] h_temp_output_;
}

float NN::testMSE()
{
    return computeMSE(h_test_inputs_, h_test_targets_);
}

float NN::testTPR(int *P, int *N, int *TP, int *FP, int *TN, int *FN)
{
    int positives = 0, negatives = 0;
    int truePositives = 0, falsePositives = 0, trueNegative = 0, falseNegative = 0;

    int numSamples = getNumInputSamples(h_test_inputs_);
    for(int i=0;i<numSamples;i++)
    {
        // calculate output
        simulateNeuralNetwork(&h_test_inputs_[i*getInputSize()], &h_weights_[0], h_output_);

        // true positive rate
        for(int j=0;j<getOutputSize();j++)
        {
            int output = classify(h_output_[j]);
            int target = classify(h_test_targets_[i*getOutputSize()+j]);

            positives += target;
            negatives += !target;
            truePositives += output && target;
            falsePositives += output && !target;
            trueNegative += !output && !target;
            falseNegative += !output && target;
        }
    }
    cout << "positives = " << positives << "\tnegatives = " << negatives << endl;
    cout << "true positives = " << truePositives << "\tfalse positives = " << falsePositives << endl;
    cout << "true negatives = " << trueNegative << "\tfalse negatives = " << falseNegative << endl;
    if(P != NULL) *TP = positives;
    if(N != NULL) *TP = negatives;
    if(TP != NULL) *TP = truePositives;
    if(FP != NULL) *FP = falsePositives;
    if(TN != NULL) *TN = trueNegative;
    if(FN != NULL) *FN = falseNegative;
    return double(truePositives) / positives;
}

void NN::saveWeightsToFile(std::string fileName)
{
    ofstream file(fileName.c_str(), ios::out | ios::binary);
    if(!file.is_open()) error(string("Error opening file: ") + fileName);

    // write numLayers
    int numLayers = getNumLayers();
    file.write((char *)&numLayers, sizeof(int));
    // write layers
    file.write((char *)&h_layers_[0], getNumLayers() * sizeof(int));

    // write weights
    file.write((char *)&h_weights_[0], getNumWeights() * sizeof(float));

    // write input and output ranges
    file.write((char *)&in_ranges_[0], 2*getInputSize() * sizeof(float));
    file.write((char *)&h_out_ranges_[0], 2*getOutputSize() * sizeof(float));

    file.close();
}
void NN::makeLiftChart(std::string fileName)
{
    ofstream file(fileName.c_str());
    if(!file.is_open()) error(string("Error opening file: ") + fileName);

    file << "is Positive;output\n";
    for(int i=0;i<getNumTestSamples();i++)
    {
        vector<float> input;
        for(int j=0;j<getInputSize();j++)
            input.push_back(h_test_inputs_[i*getInputSize()+j]);
        file << h_test_targets_[i] << ";" << (((*this)(input))[0]) << endl;
    }
    file.close();
}


// Do not use if you value your life.
void NN::downsizeTrainingSamples(int size)
{
    if((int)h_training_inputs_.size() > size)
        h_training_inputs_.resize(getInputSize() * size);
    if((int)h_training_targets_.size() > size)
        h_training_targets_.resize(getOutputSize() * size);
    // device mem will still be allocated, but those samples will not be used
}

int NN::getInputSize()
{
    return h_layers_[0];
}
int NN::getOutputSize()
{
    return h_layers_.back();
}
int NN::getLayerSize(int layer)
{
    return h_layers_[layer];
}
int NN::getNumWeights()
{
    int n = 0;
    for(int i=1;i<getNumLayers();i++)
        n += h_layers_[i] * (h_layers_[i-1]+1);

    return n;
}

int NN::getNumSynapses()
{
    return getNumWeights(); // maybe I shouldn't count the biases?
}

int NN::getNumLayers()
{
    return h_layers_.size();
}

int NN::getNumNeurons()
{
    int numNeurons=0;
    for(int i=0;i<getNumLayers();i++)
        numNeurons += h_layers_[i];
    return numNeurons;
}

int NN::getWidestLayer()
{
    int widest = 0;
    //for(int i=1;i<getNumLayers();i++)
    for(int i=0;i<getNumLayers();i++) // needs to be 0 for private mem versions
        widest = (h_layers_[i]>widest)?h_layers_[i]:widest;
    return widest;
}

int NN::getNumTrainingSamples()
{
    return getNumInputSamples(h_training_inputs_);
}

int NN::getNumValidationSamples()
{
    return getNumInputSamples(h_validation_inputs_);
}

int NN::getNumTestSamples()
{
    return getNumInputSamples(h_test_inputs_);
}

int NN::getNumInputSamples(const vector<float> &input)
{
    return input.size()/getInputSize();
}

int NN::getNumOutputSamples(const vector<float> &output)
{
    return output.size()/getOutputSize();
}

vector<float> NN::operator() (vector<float> input)
{
    if((int)input.size() != getInputSize())
        error("Input vector is the wrong size.");
    input = scaleInputData(input);
    return simulateNeuralNetwork(input);
}

vector<float> NN::simulateNeuralNetwork(vector<float> input)
{
    simulateNeuralNetwork(&input[0], &h_weights_[0], h_output_);
    vector<float> output(h_output_, h_output_ + getOutputSize());
    return output;
}

void NN::simulateNeuralNetwork(float *input, float *weights, float *output)
{ //ljepsa verzija je sporija
    int num_layers = getNumLayers();

    if(num_layers%2==0) propagateLayer(input, weights, output,h_layers_[0],h_layers_[1]);
    else propagateLayer(input, weights, h_temp_output_,h_layers_[0],h_layers_[1]);
    int weights_index = h_layers_[1]*(h_layers_[0]+1);

    for(int i=2;i<num_layers;i++)
    {
        if((num_layers-i)%2==1) propagateLayer(h_temp_output_, &weights[weights_index], output,h_layers_[i-1],h_layers_[i]);
        else propagateLayer(output, &weights[weights_index], h_temp_output_,h_layers_[i-1],h_layers_[i]);
        weights_index += h_layers_[i]*(h_layers_[i-1]+1);
    }

    // now scale the output from [0,1]
    int outputSize = getOutputSize();
    for(int i=0;i<outputSize;i++)
        output[i] = output[i] * (h_out_ranges_[2*i+1] - h_out_ranges_[2*i]) + h_out_ranges_[2*i];

    //result is now in output
}

void NN::propagateLayer(float *input, float *weights, float *output,int previous_layer_size,int layer_size)//, int isLastLayer)
{
    for(int i=0;i<layer_size;i++)
    {
        float temp_output = weights[i*(previous_layer_size+1)];
        for(int j=0;j<previous_layer_size;j++)
        {
            temp_output += input[j]*weights[i*(previous_layer_size+1)+j+1];
        }
        temp_output = sigmoid(temp_output);
        output[i] = temp_output;
    }
}

float NN::computeMSE(vector<float> &input, vector<float> &target)
{
    double squaredError=0;
    int numSamples = getNumInputSamples(input);

    for(int i=0;i<numSamples;i++)
    {
        // calculate output
        simulateNeuralNetwork(&input[i*getInputSize()], &h_weights_[0], h_output_);

        // calculate squared error
        for(int j=0;j<getOutputSize();j++)
        {
            double err = target[i*getOutputSize()+j] - h_output_[j];
            squaredError += err * err;
        }
    }
    return squaredError / (numSamples * getOutputSize());
}

void NN::printInfo()
{
    cout << "Neural network info" << endl << "-------------------" << endl;
    cout << "Number of training samples: " << getNumTrainingSamples() << endl;
    cout << "Number of validation samples: " << getNumValidationSamples() << endl;
    cout << "Number of test samples: " << getNumTestSamples() << endl;
    cout << "Layers: ";
    for(int i=0;i<getNumLayers();i++) cout << h_layers_[i] << " ";
    cout << "\nInput ranges:\n";
    for(int i=0;i<getInputSize();i++) cout << "[" << in_ranges_[2*i] << "," << in_ranges_[2*i+1] << "]" << endl;
    cout << "Output ranges:\n";
    for(int i=0;i<getOutputSize();i++) cout << "[" << h_out_ranges_[2*i] << "," << h_out_ranges_[2*i+1] << "]" << endl;
    cout << "Total weights: " << getNumWeights() << endl;
}

// NN_PSO

NN_PSO::NN_PSO(string nnetConfigFile, string trainingDataFile, string psoConfigFile) :
    NN(nnetConfigFile, trainingDataFile)
{
    if(!psoConfigFile.empty()) psoParameters_ = PsoParameters(psoConfigFile);
    psoParameters_.nDimensions = getNumWeights();

    // allocate host memory
    h_weights_ = vector<float>(getNumWeights()*getNumberOfNetworks());
    h_outputs_ = vector<float>(getOutputSize() * max(max(getNumTrainingSamples(),getNumValidationSamples()),getNumTestSamples())*getNumberOfNetworks());
}

NN_PSO::~NN_PSO()
{
}

int NN_PSO::getNumberOfNetworks()
{
    return psoParameters_.populationSize;
}

int NN_PSO::getNumSimulations()
{
    return getNumberOfNetworks()*getNumTrainingSamples();
}


void NN_PSO::simulateNeuralNetworks(vector<float> &input)
{
    Timer timer1,timer2,timer3;
    timer1.disable();
    timer2.disable();
    timer2.start();
    timer1.start();
    if(getNumInputSamples(input)*getNumberOfNetworks() > getNumOutputSamples(h_outputs_))
    {
        h_outputs_ = vector<float>(getNumInputSamples(input)*getOutputSize()*getNumberOfNetworks());
    }

    // ako se direktno koriste vectori i class varijable, onda je 2.5 puta sporije
    float *weights = &h_weights_[0];
    float *inputs = &input[0];
    int number_of_inputs = getNumInputSamples(input);
    int *layers = &h_layers_[0];
    float *output = h_output_;
    int output_size = getOutputSize();
    int num_weights = getNumWeights();
    int number_of_networks_ = getNumberOfNetworks();

    timer1.print("setup");

    for(int i=0;i<number_of_networks_;i++)
    {
        for(int j=0;j<number_of_inputs;j++)
        {
            simulateNeuralNetwork(&inputs[j*layers[0]], &weights[num_weights*i], output);
            std::copy(output, output+output_size,&h_outputs_[i*number_of_inputs*output_size+output_size*j]);
        }
    }

    timer1.print("main calc");
    timer2.print("total");
}

double NN_PSO::train()
{
    if(psoParameters_.maxFitness == 0 && psoParameters_.iterations == 0 && psoParameters_.time == 0)
        error("No terminating conditions",-79);

    // for end condtions
    Timer trainTimer;
    trainTimer.start();

    // for performance timings
    Timer performanceTimer;

    Timer timer1, timer2;
    if(!PRINT_DETAIL_TIMING_INFORMATION) { timer1.disable(); timer2.disable(); }

    ParticleSwarmOptimization pso(psoParameters_,h_weights_);
    float validationMSE = numeric_limits<float>::max();

    float bestMSE = numeric_limits<float>::max();
    vector<float> bestWeights(getNumWeights(),0);

    int iteration=0;


    float trainingMSE;

    while( ((psoParameters_.maxFitness == 0)?(1):(bestMSE > psoParameters_.maxFitness)) &&
           ((psoParameters_.iterations == 0)?(1):(iteration < psoParameters_.iterations)) &&
           ((psoParameters_.time == 0)?(1):(trainTimer.elapsed() <  psoParameters_.time)) )
    {
        if(iteration==0) performanceTimer.start(); // doesn't count preprocesing


        timer2.start();

        timer1.start();
        simulateNeuralNetworks(h_training_inputs_);
        timer1.stop();
        timer1.print("sim");

        timer1.start();
        vector<float> mses = computeMSEs(h_training_targets_);
        timer1.stop();
        timer1.print("mses");

        timer1.start();
        pso.updateFitnesses(mses);
        timer1.stop();
        timer1.print("updateFitnesses");

        timer1.start();
        pso.nextIteration();
        timer1.stop();
        timer1.print("next iter");

        timer2.stop();
        timer2.print("Total");


        iteration++;


        if(iteration % psoParameters_.stripLength == 0  && getNumValidationSamples() > 0)
        {
            simulateNeuralNetworks(h_validation_inputs_);
            vector<float> validationMSEs = computeMSEs(h_validation_targets_);
            vector<float>::iterator bestMSEit = min_element(validationMSEs.begin(), validationMSEs.end());
            validationMSE = *bestMSEit;
            if(validationMSE < bestMSE)
            {
                bestParticleIndex = bestMSEit - validationMSEs.begin();
                int start = bestParticleIndex * getNumWeights();
                int end = start + getNumWeights();
                copy(h_weights_.begin() + start, h_weights_.begin() + end ,bestWeights.begin());
                bestMSE = validationMSE;
            }
        }


        if(iteration%PRINT_GAP==0 && PRINT_INFORMATION)
        {
            simulateNeuralNetworks(h_training_inputs_);
            vector<float> trainingMSEs = computeMSEs(h_training_targets_);
            trainingMSE = *min_element(trainingMSEs.begin(), trainingMSEs.end());
            cout << "training mse[" << iteration << "]= " << trainingMSE << "\t validation mse = " << validationMSE << "\t best mse = " << bestMSE << endl;

        }

    }

    performanceTimer.stop();

    // best particle is in the first place
    copy(bestWeights.begin(), bestWeights.end(),h_weights_.begin());


    return performanceTimer.elapsed();
}

// Computes mean squared error between h_outputs and targets
vector<float> NN_PSO::computeMSEs(const vector<float> &targets)
{
    int number_of_networks = getNumberOfNetworks();
    vector<float> nMSE(number_of_networks);
    int output_size = getOutputSize();
    int num_samples = getNumOutputSamples(targets);
    for(int i = 0; i < number_of_networks; i++)
    {
        float mse = 0;
        for(int j = 0; j < num_samples; j++)
        {
            for(int k = 0; k < output_size;k++)
            {
                float d = h_outputs_[(i*num_samples+j)*output_size+k] - targets[j*output_size+k];
                mse += d * d;
            }
        }
        nMSE[i] = mse/(num_samples*output_size);
    }
    return nMSE;
}

void NN_PSO::printInfo()
{
    NN::printInfo();
    psoParameters_.PrintInfo();
}

double NN_PSO::timing(int numIterations)
{
    float maxFitness = 0;
    float time = 0;
    swap(psoParameters_.iterations, numIterations);
    swap(psoParameters_.maxFitness, maxFitness);
    swap(psoParameters_.time, time);

    double elapsed = train();

    // restore to previous conditions
    swap(psoParameters_.iterations, numIterations);
    swap(psoParameters_.maxFitness, maxFitness);
    swap(psoParameters_.time, time);

    return elapsed / numIterations;
}


// NN_OpenCL

NN_OpenCL::NN_OpenCL(string nnetConfigFile, string trainingDataFile, cl_device_type device_type) :
    NN(nnetConfigFile,trainingDataFile), device_type_(device_type)
{
    cl_int err;

    // create context
    context_ = openclCreateContextFromType(device_type_,&device_);

    // create command queue
    queue_ = clCreateCommandQueue(context_,device_,0,&err);
    if(err != CL_SUCCESS) error("Error creating queue", 553);

    // allocate memory and copy data to device
    cl_int err_a[100]; int err_count=0;
    d_training_inputs_ = clCreateBuffer(context_,CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, h_training_inputs_.size()*sizeof(float),&h_training_inputs_[0],&err_a[err_count++]);
    d_training_targets_ = clCreateBuffer(context_,CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, h_training_targets_.size()*sizeof(float),&h_training_targets_[0],&err_a[err_count++]);
    if(getNumValidationSamples() > 0) // size == 0 is invalid buffer size
    {
        d_validation_inputs_ = clCreateBuffer(context_,CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, h_validation_inputs_.size()*sizeof(float),&h_validation_inputs_[0],&err_a[err_count++]);
        d_validation_targets_ = clCreateBuffer(context_,CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, h_validation_targets_.size()*sizeof(float),&h_validation_targets_[0],&err_a[err_count++]);
    }
    if(getNumTestSamples() > 0) // size == 0 is invalid buffer size
    {
        d_test_inputs_ = clCreateBuffer(context_,CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, h_test_inputs_.size()*sizeof(float),&h_test_inputs_[0],&err_a[err_count++]);
        d_test_targets_ = clCreateBuffer(context_,CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, h_test_targets_.size()*sizeof(float),&h_test_targets_[0],&err_a[err_count++]);
    }

    d_out_ranges_ = clCreateBuffer(context_,CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, h_out_ranges_.size()*sizeof(float),&h_out_ranges_[0],&err_a[err_count++]);
    d_layers_ = clCreateBuffer(context_,CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, h_layers_.size()*sizeof(int),&h_layers_[0],&err_a[err_count++]);
    d_weights_ = clCreateBuffer(context_,CL_MEM_READ_ONLY, h_weights_.size()*sizeof(float),NULL,&err_a[err_count++]);

    for(int i=0;i<err_count;i++)
        if(err_a[i] != CL_SUCCESS) error("Error creating buffer",636);

    // create program
    program_ = openclCreateProgramFromFile(context_, KERNEL_FILE_NAME);

    // build program
    openclBuildProgram(program_, device_);

}
void NN_OpenCL::printInfo()
{
    cout << "Device info\n--------------------" << endl;
    openclPrintDeviceInfo(device_);
}

NN_OpenCL::~NN_OpenCL()
{
    clReleaseProgram(program_);

    clReleaseMemObject(d_out_ranges_);
    clReleaseMemObject(d_layers_);
    clReleaseMemObject(d_weights_);
    clReleaseMemObject(d_outputs_);

    clReleaseMemObject(d_training_inputs_);
    clReleaseMemObject(d_training_targets_);
    if(getNumValidationSamples() > 0)
    {
        clReleaseMemObject(d_validation_targets_);
        clReleaseMemObject(d_validation_inputs_);
    }
    if(getNumTestSamples() > 0)
    {
        clReleaseMemObject(d_test_targets_);
        clReleaseMemObject(d_test_inputs_);
    }

    clReleaseCommandQueue(queue_);
    clReleaseContext(context_);
}

// returns a cl_mem object (on device) which is equivalent to the vector (on host)
cl_mem NN_OpenCL::vectorToClmem(const vector<float> &vec)
{
    cl_mem d_vec=d_training_inputs_;
    if(&vec == &h_training_inputs_)
        d_vec = d_training_inputs_;
    else if(&vec == &h_training_targets_)
        d_vec = d_training_targets_;
    else if(&vec == &h_validation_inputs_)
        d_vec = d_validation_inputs_;
    else if(&vec == &h_validation_targets_)
        d_vec = d_validation_targets_;
    else if(&vec == &h_test_inputs_)
        d_vec = d_test_inputs_;
    else if(&vec == &h_test_targets_)
        d_vec = d_test_targets_;
    else error("Only training, validation and test samples supported for now.", 395);
    return d_vec;
}

// NN_PSO_OpenCL

NN_PSO_OpenCL::NN_PSO_OpenCL(string nnetConfigFile, string trainingDataFile, string psoConfigFile, cl_device_type device_type) :
    NN(nnetConfigFile,trainingDataFile), NN_PSO(nnetConfigFile,trainingDataFile,psoConfigFile), NN_OpenCL(nnetConfigFile,trainingDataFile,device_type)
{
    cl_int err;
    // allocate memory on device
    cl_int err_a[100]; int err_count=0;

    d_outputs_ = clCreateBuffer(context_,CL_MEM_READ_WRITE, h_outputs_.size()*sizeof(float),NULL,&err_a[err_count++]);
    d_MSEs_ = clCreateBuffer(context_,CL_MEM_WRITE_ONLY, getNumberOfNetworks()*sizeof(float),NULL,&err_a[err_count++]);

    for(int i=0;i<err_count;i++)
        if(err_a[i] != CL_SUCCESS) error("Error creating buffers",639);


    // create MSE kernel
    mse_kernel_ = clCreateKernel(program_, "mean_squared_errors", &err);
    if(err != CL_SUCCESS) error("Error creating MSE kernel", 873);

    // set kernel arguments
    err =clSetKernelArg(mse_kernel_, 0, sizeof(cl_mem), &d_outputs_);
    err|=clSetKernelArg(mse_kernel_, 1, sizeof(cl_mem), &d_training_targets_);
    err|=clSetKernelArg(mse_kernel_, 2, sizeof(cl_mem), &d_MSEs_);
    int outputSize = getOutputSize();
    err|=clSetKernelArg(mse_kernel_, 4, sizeof(int), &outputSize);
    mseGroupSize_ = 64; // not really that important
    err|=clSetKernelArg(mse_kernel_, 5, mseGroupSize_ * sizeof(float), NULL);

    // arguments 1, 3 will have to be redefined later
}

void NN_PSO_OpenCL::printInfo()
{
    NN_PSO::printInfo();
    cout << "Group size: " << simGroupSize_ << endl;
    NN_OpenCL::printInfo();
}


NN_PSO_OpenCL::~NN_PSO_OpenCL()
{
    clReleaseKernel(mse_kernel_);
    clReleaseMemObject(d_MSEs_);
}

// Calculates outputs for the given inputs for each neural network.
// Transfers the weights to device, runs the kernel.
// Output stays on device where mse will continue with calculations.
// For now, only acceptable inputs are training and validation samples (which are already on device).
// Inputs and input sizes have to be parameters 0 and 7 for any kernel.
void NN_PSO_OpenCL::simulateNeuralNetworks(vector<float> &inputs)
{
    cl_int err = 0;

    cl_mem d_inputSamples = vectorToClmem(inputs);
    cl_int numSamples = getNumInputSamples(inputs);

    err = clSetKernelArg(sim_kernel_, 0, sizeof(cl_mem), &d_inputSamples);
    err |= clSetKernelArg(sim_kernel_, 7, sizeof(cl_int), &numSamples);
    if(err != CL_SUCCESS) error("Error setting arguments in simNNs",151);


    // write weights to device
    err = clEnqueueWriteBuffer(queue_, d_weights_, CL_FALSE, 0, h_weights_.size()*sizeof(float), &h_weights_[0], 0, NULL, NULL);
    if(err != CL_SUCCESS) error("Error writing weights to buffer",152);

    // work-group sizes
    size_t localWorkSize[2];
    localWorkSize[0] = 1;
    localWorkSize[1] = simGroupSize_;
    size_t globalWorkSize[2];
    globalWorkSize[0] = getNumberOfNetworks();
    int numGroupsY = (numSamples + simGroupSize_ - 1)/simGroupSize_;
    globalWorkSize[1] = numGroupsY * simGroupSize_;

    err = clEnqueueNDRangeKernel(queue_, sim_kernel_, 2, NULL, globalWorkSize,localWorkSize, 0, NULL,NULL);

    if(err != CL_SUCCESS)
    {
        cout << "kernel err= " << err << endl;
        error("Error running kernel ", 504);
    }

    CL_FINISH // for timing purposes

    // results will stay on device for mse
    // copy back results
    //err = clEnqueueReadBuffer(queue_, d_outputs_, CL_TRUE, 0, getNumberOfNetworks()*numSamples*getOutputSize()*sizeof(float), &h_outputs_[0], 0, NULL, NULL);
    //if(err != CL_SUCCESS) error("Error reading back outputs buffer", 671);
}

vector<float> NN_PSO_OpenCL::computeMSEs(const vector<float> &targets)
{
    vector<float> nMSE(getNumberOfNetworks());

    cl_int err=0;
    // outputs are already on device

    // set kernel args (pointers to targets and num samples)
    cl_mem d_targets = vectorToClmem(targets);
    cl_int numSamples = getNumOutputSamples(targets);
    err=clSetKernelArg(mse_kernel_, 1, sizeof(cl_mem), &d_targets);
    err|=clSetKernelArg(mse_kernel_, 3, sizeof(cl_int), &numSamples);
    if(err != CL_SUCCESS) error("Error seting kernel arguments in mse", 366);

    // run kernel
    size_t localWorkSize = mseGroupSize_;
    size_t globalWorkSize = localWorkSize * getNumberOfNetworks();
    err = clEnqueueNDRangeKernel(queue_, mse_kernel_, 1, NULL, &globalWorkSize,&localWorkSize, 0, NULL,NULL);
    if(err != CL_SUCCESS) error("Error running mse kernel", 466);

    // copy back MSEs;
    err = clEnqueueReadBuffer(queue_, d_MSEs_, CL_TRUE, 0, getNumberOfNetworks()*sizeof(float), &nMSE[0], 0, NULL, NULL);
    if(err != CL_SUCCESS) error("Error reading back mses", 670);

    return nMSE;
}

// local memory parameters must be unset.
int NN_PSO_OpenCL_local::getMaxWorkGroupSize()
{
    // maximum size based on register usage
    size_t grpSize;
    int err = clGetKernelWorkGroupInfo(sim_kernel_,device_, CL_KERNEL_WORK_GROUP_SIZE, sizeof(size_t), &grpSize, NULL);

    // size of local memory used for parameter passing and such
    cl_ulong locSize;
    err |= clGetKernelWorkGroupInfo(sim_kernel_,device_, CL_KERNEL_LOCAL_MEM_SIZE , sizeof(cl_ulong), &locSize, NULL);

    // total local memory for the device
    cl_ulong maxLocSize;
    err |= clGetDeviceInfo(device_, CL_DEVICE_LOCAL_MEM_SIZE, sizeof(cl_ulong), &maxLocSize, NULL);

    if(err != CL_SUCCESS) error("Error finding max group size",843);

    // memory left for individual work-items/threads (there is only one set of weights per block)
    int marginOfError = 256; // if the device uses local memory for some other things it doesn't report
    int memLeft = maxLocSize - locSize - getNumWeights() * sizeof(float) - marginOfError;

    int memPerThread = 2 * getWidestLayer() * sizeof(float); // for output and temp_output buffers

    // group size based on local memory;
    int groupSize = memLeft / memPerThread;

    // group size based on local memory and other limits (registers, max block size etc.)
    groupSize = min(groupSize, (int)grpSize);

    return groupSize;
}


NN_PSO_OpenCL_local::NN_PSO_OpenCL_local(string nnetConfigFile, string trainingDataFile, string psoConfigFile, cl_device_type device_type) :
    NN(nnetConfigFile,trainingDataFile), NN_PSO_OpenCL(nnetConfigFile,trainingDataFile,psoConfigFile,device_type)
{
    cl_int err;

    // kernel
    sim_kernel_ = clCreateKernel(program_, "simulate_neural_networks_local", &err);
    if(err != CL_SUCCESS) error("Error creating kernel", 897);

    // set kernel arguments
    err=clSetKernelArg(sim_kernel_, 1, sizeof(cl_mem), &d_weights_);
    err|=clSetKernelArg(sim_kernel_, 2, sizeof(cl_mem), &d_outputs_);
    int numLayers = getNumLayers();
    err|=clSetKernelArg(sim_kernel_, 3, sizeof(int), &numLayers);
    err|=clSetKernelArg(sim_kernel_, 4, sizeof(cl_mem), &d_layers_);
    err|=clSetKernelArg(sim_kernel_, 5, sizeof(cl_mem), &d_out_ranges_);
    int numberOfNetworks = getNumberOfNetworks(); // maybe remove this?
    err|=clSetKernelArg(sim_kernel_, 6, sizeof(int), &numberOfNetworks);
    int numWeights = getNumWeights();
    err|=clSetKernelArg(sim_kernel_, 8, sizeof(int), &numWeights);
    int widestLayer = getWidestLayer();
    err|=clSetKernelArg(sim_kernel_, 9, sizeof(int), &widestLayer);
    err|=clSetKernelArg(sim_kernel_, 10, getNumWeights() * sizeof(float), NULL);

    simGroupSize_ = min(getMaxWorkGroupSize(), 64);

    // arguments 0, 7 will have to be redefined later

    // arguments 11,12 (local memory size) depend on work group size

    err|=clSetKernelArg(sim_kernel_, 11, simGroupSize_ * getWidestLayer() * sizeof(float), NULL);
    err|=clSetKernelArg(sim_kernel_, 12, simGroupSize_ * getWidestLayer() * sizeof(float), NULL);

    if(err!=CL_SUCCESS) error("Error setting kernel arguments",548);
}

NN_PSO_OpenCL_local::~NN_PSO_OpenCL_local()
{
    clReleaseKernel(sim_kernel_);
}

void NN_PSO_OpenCL_local::printInfo()
{
    NN_PSO::printInfo();
    cout << "Work-group size: " << simGroupSize_ << endl;
    cout << "Using local memory." << endl;
    NN_OpenCL::printInfo();
}


NN_PSO_OpenCL_GPU_local::NN_PSO_OpenCL_GPU_local(string nnetConfigFile, string trainingDataFile, string psoConfigFile) :
    NN(nnetConfigFile,trainingDataFile), NN_PSO_OpenCL_local(nnetConfigFile,trainingDataFile,psoConfigFile, CL_DEVICE_TYPE_GPU)
{
}

NN_PSO_OpenCL_CPU_local::NN_PSO_OpenCL_CPU_local(string nnetConfigFile, string trainingDataFile, string psoConfigFile) :
    NN(nnetConfigFile,trainingDataFile), NN_PSO_OpenCL_local(nnetConfigFile,trainingDataFile,psoConfigFile, CL_DEVICE_TYPE_CPU)
{
}


NN_PSO_OpenCL_private::NN_PSO_OpenCL_private(string nnetConfigFile, string trainingDataFile, string psoConfigFile, cl_device_type device_type) :
    NN(nnetConfigFile,trainingDataFile), NN_PSO_OpenCL(nnetConfigFile,trainingDataFile,psoConfigFile,device_type)
{
    cl_int err;

    // release the previous unedited version
    clReleaseProgram(program_);

    // read in source and make changes
    string source = readFile(KERNEL_FILE_NAME);
    stringstream arraySize;
    arraySize << "[" << getWidestLayer() << "]";
    source = findAndReplace(source, "[PRIVATE_MEM_SIZE]", arraySize.str());

    // create program
    program_ = openclCreateProgramFromSource(context_, source);

    // build program
    openclBuildProgram(program_, device_);

    // kernel
    sim_kernel_ = clCreateKernel(program_, "simulate_neural_networks_private", &err);
    if(err != CL_SUCCESS) error("Error creating kernel", 897);

    // set kernel arguments
    err=clSetKernelArg(sim_kernel_, 0, sizeof(cl_mem), &d_training_inputs_);
    err|=clSetKernelArg(sim_kernel_, 1, sizeof(cl_mem), &d_weights_);
    err|=clSetKernelArg(sim_kernel_, 2, sizeof(cl_mem), &d_outputs_);
    int numLayers = getNumLayers();
    err|=clSetKernelArg(sim_kernel_, 3, sizeof(int), &numLayers);
    err|=clSetKernelArg(sim_kernel_, 4, sizeof(cl_mem), &d_layers_);
    err|=clSetKernelArg(sim_kernel_, 5, sizeof(cl_mem), &d_out_ranges_);
    int numberOfNetworks = getNumberOfNetworks(); // maybe remove this?
    err|=clSetKernelArg(sim_kernel_, 6, sizeof(int), &numberOfNetworks);
    int numInputs = getNumTrainingSamples();
    err|=clSetKernelArg(sim_kernel_, 7, sizeof(int), &numInputs);
    int numWeights = getNumWeights();
    err|=clSetKernelArg(sim_kernel_, 8, sizeof(int), &numWeights);

    if(err!=CL_SUCCESS) error("Error setting kernel arguments",548);

    simGroupSize_ = min(getMaxWorkGroupSize(), 64);

    // arguments 0, 7 will have to be redefined later
}

NN_PSO_OpenCL_private::~NN_PSO_OpenCL_private()
{
    clReleaseKernel(sim_kernel_);
}

int NN_PSO_OpenCL_private::getMaxWorkGroupSize()
{
    // maximum size based on register usage, etc.
    size_t groupSize;
    int err = clGetKernelWorkGroupInfo(sim_kernel_,device_, CL_KERNEL_WORK_GROUP_SIZE, sizeof(size_t), &groupSize, NULL);
    if(err != CL_SUCCESS) error("Error finding max group size",844);

    return groupSize;
}

void NN_PSO_OpenCL_private::printInfo()
{
    NN_PSO::printInfo();
    cout << "Work-group size: " << simGroupSize_ << endl;
    cout << "Using private memory." << endl;
    NN_OpenCL::printInfo();
}

NN_PSO_OpenCL_GPU_private::NN_PSO_OpenCL_GPU_private(string nnetConfigFile, string trainingDataFile, string psoConfigFile) :
    NN(nnetConfigFile,trainingDataFile), NN_PSO_OpenCL_private(nnetConfigFile,trainingDataFile,psoConfigFile, CL_DEVICE_TYPE_GPU)
{
}

NN_PSO_OpenCL_CPU_private::NN_PSO_OpenCL_CPU_private(string nnetConfigFile, string trainingDataFile, string psoConfigFile) :
    NN(nnetConfigFile,trainingDataFile), NN_PSO_OpenCL_private(nnetConfigFile,trainingDataFile,psoConfigFile, CL_DEVICE_TYPE_CPU)
{
}


// Back propagation

// BpParameters

BpParameters::BpParameters()
{
    SetDefaults();
}

BpParameters::BpParameters(string fileName)
{
    SetDefaults();

    ifstream file(fileName.c_str(), ios::in);
    if(!file.is_open()) error(string("Error opening file: ") + fileName, -238);


    bool terminatingConditions = false; // were there terminating conditions before
    char line[101];
    while(file.getline(line,100))
    {
        //string line(cline);
        double result;

        if(ParseLine(line, "Learning rate", result))
            learningRate = result;

        if(ParseLine(line, "Strip length", result))
            stripLength = result;

        if(ParseLine(line, "Lower limit", result))
            lowerLimit = result;

        if(ParseLine(line, "Upper limit", result))
            upperLimit = result;

        // Terminating conditions
        if(ParseLine(line, "Iterations", result))
        {
            if(terminatingConditions == false) // this is the first
            {
                time = 0;
                maxFitness = 0;
                terminatingConditions = true;
            }
            iterations = result;
        }

        if(ParseLine(line, "Time", result))
        {
            if(terminatingConditions == false) // this is the first
            {
                iterations = 0;
                maxFitness = 0;
                terminatingConditions = true;
            }
            time = result;
        }

        if(ParseLine(line, "MSE", result))
        {
            if(terminatingConditions == false) // this is the first
            {
                iterations = 0;
                time = 0;
                terminatingConditions = true;
            }
            maxFitness = result;
        }

    }
    file.close();

}

void BpParameters::SetDefaults()
{
    learningRate = 0.02;

    stripLength = 10;

    lowerLimit = -1;
    upperLimit = 1;

    iterations = 0;
    maxFitness = 0;
    time = 60;
}
void BpParameters::PrintInfo()
{
    cout << "BP parameters info:\n";
    cout << "  Learn rate: " << learningRate << endl;
    cout << "  Strip length: " << stripLength << endl;
    cout << "  Limits : [" << lowerLimit << ", " << upperLimit << "]\n";
    cout << "  Time: " << time << endl;
    cout << "  Iterations: " << iterations << endl;
    cout << "  MSE: " << maxFitness << endl;
}



// NN_BP
NN_BP::NN_BP(string nnetConfigFile, string trainingDataFile, string bpConfigFile) :
    NN(nnetConfigFile, trainingDataFile)
{
    if(!bpConfigFile.empty()) bpParameters_ = BpParameters(bpConfigFile);

    // allocate host memory
    h_weights_ = vector<float>(getNumWeights());
    h_outputs_ = vector<float>(getWidestLayer());

    // allocate memory for CPU simulation
    h_deltas_ = new float[getWidestLayer()];
    h_prev_deltas_ = new float[getWidestLayer()];
    h_all_outputs_ = new float[getNumNeurons()];
    h_delta_weights_ = new float[getNumWeights()];
}

NN_BP::~NN_BP()
{
    delete [] h_deltas_;
    delete [] h_prev_deltas_;
    delete [] h_all_outputs_;
    delete [] h_delta_weights_;
}

void NN_BP::simulateNeuralNetworkForBP(float *input, float *weights, float *output, float *all_outputs)
{
    int neuron_index = 0;

    // output from first neurons == input
    for(int j=0;j<h_layers_[0];j++) all_outputs[neuron_index++] = input[j];

    if(getNumLayers()%2==0)
    {
        propagateLayer(input, weights, output,h_layers_[0],h_layers_[1]);
        for(int j=0;j<h_layers_[1];j++) all_outputs[neuron_index++] = output[j];
    }
    else
    {
        propagateLayer(input, weights, h_temp_output_,h_layers_[0],h_layers_[1]);
        for(int j=0;j<h_layers_[1];j++) all_outputs[neuron_index++] = h_temp_output_[j];
    }
    int weights_index = h_layers_[1]*(h_layers_[0]+1);



    for(int i=2;i<getNumLayers();i++)
    {
        if((getNumLayers()-i)%2==1)
        {
            propagateLayer(h_temp_output_, &weights[weights_index], output,h_layers_[i-1],h_layers_[i]);
            for(int j=0;j<h_layers_[i];j++) all_outputs[neuron_index++] = output[j];
        }
        else
        {
            propagateLayer(output, &weights[weights_index], h_temp_output_,h_layers_[i-1],h_layers_[i]);
            for(int j=0;j<h_layers_[i];j++) all_outputs[neuron_index++] = h_temp_output_[j];
        }
        weights_index += h_layers_[i]*(h_layers_[i-1]+1);
    }

    // now scale the output from [0,1]
    int outputSize = getOutputSize();
    for(int i=0;i<outputSize;i++)
        output[i] = output[i] * (h_out_ranges_[2*i+1] - h_out_ranges_[2*i]) + h_out_ranges_[2*i];

    //result is now in output
}

double NN_BP::train()
{

    if(bpParameters_.maxFitness == 0 && bpParameters_.iterations == 0 && bpParameters_.time == 0)
        error("No terminating conditions",-79);

    // for end condtions
    Timer trainTimer;
    trainTimer.start();

    // for performance timings
    Timer performanceTimer;

    Timer timer1, timer2;
    if(!PRINT_DETAIL_TIMING_INFORMATION) { timer1.disable(); timer2.disable(); }

    // randomize weights
    initializeWeights();

    float bestMSE = numeric_limits<float>::max(); // MSE which is being minimized
    vector<float> bestWeights(h_weights_); // best weights so far, according to the validation samples

    double validationMSE = numeric_limits<float>::max();
    int iteration=0;

    performanceTimer.start();

    while( ((bpParameters_.maxFitness == 0)?(1):(bestMSE > bpParameters_.maxFitness)) &&
           ((bpParameters_.iterations == 0)?(1):(iteration < bpParameters_.iterations)) &&
           ((bpParameters_.time == 0)?(1):(trainTimer.elapsed() <  bpParameters_.time)) )
    {
        // one iteration is a pass through all the training samples
        for(int i=0; i<getNumTrainingSamples(); i++)
        {
            timer1.start();
            // initialize delta weights to 0
            initializeDeltaWeights();
            timer1.print("initializeDeltaWeights");

            timer1.start();
            // propagate input through the network and calculates output from all neurons.
            propagateForward(h_training_inputs_, i);
            timer1.print("propagateForward");

            timer1.start();
            // propagate errors back through the network and calculate delta weights
            propagateBackwards(h_training_targets_, i);
            timer1.print("propagateBackwards");

            timer1.start();
            // now add the delta weights to the weights after each training sample
            applyDeltaWeights();
            timer1.print("applyDeltaWeights");
        }

        // now add the delta weights to the weights after all the training samples
        //applyDeltaWeights();

        iteration++;

        if(iteration % bpParameters_.stripLength == 0 && getNumValidationSamples() > 0)
        {
            validationMSE = computeMSE(h_validation_inputs_, h_validation_targets_);

            if(validationMSE < bestMSE)
            {
                bestWeights = h_weights_;
                bestMSE = validationMSE;
            }
        }


        if(iteration%PRINT_GAP==0  && PRINT_INFORMATION)
        {
            double trainingMSE = computeMSE(h_training_inputs_,h_training_targets_);
            cout << "training mse[" << iteration << "]= " << trainingMSE << "\t validation mse = " << validationMSE << "\t best mse = " << bestMSE << endl;
        }
    }

    performanceTimer.stop();

    finalizeTraining();

    // best weights are saved
    h_weights_ = bestWeights;


    return performanceTimer.elapsed();
}

/// Sets weights to random values between lower and upper limit.
void NN_BP::initializeWeights()
{
    randomizeVector(h_weights_,bpParameters_.lowerLimit, bpParameters_.upperLimit);
}

/// Sets h_delta_weights_ to 0.
void NN_BP::initializeDeltaWeights()
{
    int numWeights = getNumWeights();
    for(int i=0;i<numWeights;i++)
        h_delta_weights_[i] = 0;
}

/// Propagates input of the sample from samples with index sampleIndex through the network
/// and sets h_all_outputs_ with with ouputs from every neurons.
void NN_BP::propagateForward(vector<float> &inputSamples, int sampleIndex)
{
    simulateNeuralNetworkForBP(&inputSamples[sampleIndex*getInputSize()], &h_weights_[0], &h_outputs_[0], h_all_outputs_);
}


void NN_BP::propagateBackwards(vector<float> &outputSamples, int sampleIndex)
{
    // scale targets to [0,1]
    vector<float> scaled_targets(outputSamples.begin() + sampleIndex*getOutputSize(), outputSamples.begin() + (sampleIndex+1)*getOutputSize());
    scaled_targets = inverseScaleOutputData(scaled_targets);


    // calculate output deltas
    float *layerOutput = getLayerOutput(h_all_outputs_, getNumLayers()-1);
    for(int j=0; j<getOutputSize();j++) // j == output neuron
    {
        h_deltas_[j] = layerOutput[j] * (1-layerOutput[j]) * (scaled_targets[j] - layerOutput[j]);
    }

    // calculate delta weights to the output neurons  //getOutputSize() == h_layers[getnumlayer-1...]
    for(int j=0; j<getOutputSize();j++) // j == output neuron
    {
        float *delta_weights = getNeuronWeights(h_delta_weights_, getNumLayers()-1, j);

        float *prev_layer_output = getLayerOutput(h_all_outputs_, getNumLayers()-2);

        // bias
        delta_weights[0] += bpParameters_.learningRate * h_deltas_[j] * 1;

        // connections to the previous layer
        for(int k=0;k<h_layers_[getNumLayers()-2];k++)
        {
            delta_weights[k+1] += bpParameters_.learningRate * h_deltas_[j] * prev_layer_output[k]; //delta_weights[0] is bias, so +1
        }
    }


    // calculate deltas and delta weights for the hidden neurons
    for(int layer=getNumLayers()-2; layer>0; layer--)
    {
        // deltas from the previous calculation
        memcpy(h_prev_deltas_, h_deltas_, h_layers_[layer+1] * sizeof(float));


        float *layerOutput = getLayerOutput(h_all_outputs_, layer);
        for(int j=0; j<h_layers_[layer];j++) // j == neuron
        {

            // calculate deltas
            float delta_sum = 0;
            for(int k=0;k<h_layers_[layer+1];k++) // k == neuron from layer+1
            {
                float *prev_neuron = getNeuronWeights(&h_weights_[0], layer+1, k); // previous in BP
                delta_sum += h_prev_deltas_[k] * prev_neuron[j+1]; // +1 becouse 0 == bias
            }
            h_deltas_[j] = layerOutput[j] * (1-layerOutput[j]) * delta_sum;


            // calculate delta weights
            float *delta_weights = getNeuronWeights(h_delta_weights_, layer, j);

            float *prev_layer_output = getLayerOutput(h_all_outputs_, layer-1); // previous in forward feedback

            // bias
            delta_weights[0] += bpParameters_.learningRate * h_deltas_[j] * 1;

            // connections to the previous layer
            for(int k=0;k<h_layers_[layer-1];k++)
            {
                delta_weights[k+1] += bpParameters_.learningRate * h_deltas_[j] * prev_layer_output[k]; //delta_weights[0] is bias, so +1
            }
        }

    }
}


/// Adds h_delta_weights_ to h_weights.
void NN_BP::applyDeltaWeights()
{
    int numWeights = getNumWeights();
    for(int j=0;j<numWeights;j++)
    {
        ///cout << "delta = " << h_delta_weights_[i] << endl;
        h_weights_[j] += h_delta_weights_[j];
    }
}

void NN_BP::finalizeTraining()
{
}


// layer and index is from 0 to N-1
float NN_BP::getOutput(float *all_outputs, int layer, int index)
{
    int total_index = 0;
    for(int i=0;i<layer;i++)
        total_index += h_layers_[i];
    total_index += index;
    return all_outputs[total_index];
}

float* NN_BP::getLayerOutput(float *all_outputs, int layer)
{
    int layer_index = 0;
    for(int i=0;i<layer;i++)
        layer_index += h_layers_[i];
    return &all_outputs[layer_index];
}

/// Returns a pointer to the weights on the link to the input of the neuron. First weight is the bias.
/// Input neurons don't have any weights, so layer [1, numLayers-1]. neuron e= [0,layerSize-1].
float* NN_BP::getNeuronWeights(float *weights, int layer, int neuron)
{
    int index = 0;

    // weights up to the layer before
    for(int i=1;i<layer;i++)
        index += h_layers_[i] * (h_layers_[i-1]+1);

    // weights up to the neuron in the layer
    index += neuron * (h_layers_[layer-1] + 1);

    return &weights[index];
}

double NN_BP::timing(int numIterations)
{
    float maxFitness = 0;
    float time = 0;
    swap(bpParameters_.iterations, numIterations);
    swap(bpParameters_.maxFitness, maxFitness);
    swap(bpParameters_.time, time);

    double elapsed = train();

    // restore to previous conditions
    swap(bpParameters_.iterations, numIterations);
    swap(bpParameters_.maxFitness, maxFitness);
    swap(bpParameters_.time, time);

    return elapsed / numIterations;
}

void NN_BP::printInfo()
{
    NN::printInfo();
    bpParameters_.PrintInfo();
}

void NN_BP::simulateNeuralNetwork(vector<float> &inputSamples, int sampleIndex)
{
    propagateForward(inputSamples, sampleIndex);
    // output is now in h_outputs_
}

float NN_BP::computeMSE(vector<float> &inputSamples, vector<float> &sampleTargets)
{
    int numSamples = getNumInputSamples(inputSamples);
    float squaredErrorSum = 0;
    for(int i=0;i<numSamples;i++)
    {
        simulateNeuralNetwork(inputSamples, i); // output is in h_outputs_
        for(int j=0;j<getOutputSize();j++)
        {
            double error = h_outputs_[j] - sampleTargets[i*getOutputSize()+j];
            squaredErrorSum += error * error;
        }
    }
    return squaredErrorSum / sampleTargets.size();
}

// NN_BP_OpenCL

NN_BP_OpenCL::NN_BP_OpenCL(string nnetConfigFile, string trainingDataFile, string bpConfigFile, cl_device_type device_type):
    NN(nnetConfigFile,trainingDataFile), NN_BP(nnetConfigFile,trainingDataFile,bpConfigFile), NN_OpenCL(nnetConfigFile,trainingDataFile,device_type)
{
    cl_int err;

    // allocate memory on device
    cl_int err_a[100]; int err_count=0;
    d_all_outputs_ = clCreateBuffer(context_,CL_MEM_READ_WRITE,   getNumNeurons()*sizeof(float),NULL,&err_a[err_count++]);
    d_prev_deltas_ = clCreateBuffer(context_,CL_MEM_READ_WRITE,   getWidestLayer()*sizeof(float),NULL,&err_a[err_count++]);
    d_deltas_ = clCreateBuffer(context_,CL_MEM_READ_WRITE,   getWidestLayer()*sizeof(float),NULL,&err_a[err_count++]);
    d_delta_weights_ = clCreateBuffer(context_,CL_MEM_READ_WRITE,   getNumWeights()*sizeof(float),NULL,&err_a[err_count++]);
    d_outputs_ = clCreateBuffer(context_,CL_MEM_READ_WRITE,   16*sizeof(float),NULL,&err_a[err_count++]); // token buffer, for NN_OpenCL destructor (should refactor, who creates it, releases it)
    d_output_ = clCreateBuffer(context_,CL_MEM_READ_WRITE,   getWidestLayer()*sizeof(float),NULL,&err_a[err_count++]);
    d_temp_output_ = clCreateBuffer(context_,CL_MEM_READ_WRITE,   getWidestLayer()*sizeof(float),NULL,&err_a[err_count++]);

    for(int i=0;i<err_count;i++)
        if(err_a[i] != CL_SUCCESS) error("Error creating buffers",638);

    // ----- Create kernels -----
    // create layer_propagate_forward kernel
    layer_propagate_forward_kernel_ = clCreateKernel(program_, "layer_propagate_forward", &err);
    if(err != CL_SUCCESS) error("Error creating layer_propagate_forward kernel", 833);
    // set kernel arguments
    err = clSetKernelArg(layer_propagate_forward_kernel_, 2, sizeof(cl_mem), &d_weights_);
    err|= clSetKernelArg(layer_propagate_forward_kernel_, 5, sizeof(cl_mem), &d_all_outputs_);
    err|= clSetKernelArg(layer_propagate_forward_kernel_, 10, sizeof(cl_mem), &d_out_ranges_);
    // arguments 0,1,3,4,6,7,8,9 will have to be set while trainning
    if(err != CL_SUCCESS) error("Error setting kernel arguments for propagate layer forward kernel", 547);

    // create layer_propagate_backwards kernel
    layer_propagate_backwards_kernel_ = clCreateKernel(program_, "layer_propagate_backwards", &err);
    if(err != CL_SUCCESS) error("Error creating layer_propagate_backwards kernel", 233);
    // set kernel arguments
    err = clSetKernelArg(layer_propagate_backwards_kernel_, 0, sizeof(cl_mem), &d_all_outputs_);
    // 1 all_outputs_offset
    // 2 prev_deltas
    // 3 deltas
    err|= clSetKernelArg(layer_propagate_backwards_kernel_, 4, sizeof(cl_mem), &d_weights_);
    // 5 weights_offset
    err|= clSetKernelArg(layer_propagate_backwards_kernel_, 6, sizeof(cl_mem), &d_delta_weights_);
    // 7 prev_layer_size
    // 8 layer_size
    // 9 next_layer_size
    err|= clSetKernelArg(layer_propagate_backwards_kernel_, 10, sizeof(float), &(bpParameters_.learningRate));
    // 11 isLastLayer
    err|= clSetKernelArg(layer_propagate_backwards_kernel_, 12, sizeof(cl_mem), &d_training_targets_);
    // 13 targets_offset
    err|= clSetKernelArg(layer_propagate_backwards_kernel_, 14, sizeof(cl_mem), &d_out_ranges_);
    // arguments 1,5,7,8,9,11,13 will have to be set while training
    if(err != CL_SUCCESS) error("Error setting kernel arguments for propagate layer backwards kernel", 547);

    // create cl_kernel apply_delta_weights kernel
    apply_delta_weights_kernel_ =  clCreateKernel(program_, "apply_delta_weights", &err);
    if(err != CL_SUCCESS) error("Error creating apply_delta_weights kernel", 222);
    err = clSetKernelArg(apply_delta_weights_kernel_, 0, sizeof(float), &d_weights_);
    err|= clSetKernelArg(apply_delta_weights_kernel_, 1, sizeof(float), &d_delta_weights_);
    cl_int num_weights = getNumWeights();
    err|= clSetKernelArg(apply_delta_weights_kernel_, 2, sizeof(cl_int), &num_weights);
    if(err != CL_SUCCESS) error("Error setting kernel arguments for apply_delta_weights kernel", 547);
}

NN_BP_OpenCL::~NN_BP_OpenCL()
{
    // Release memory objects.
    clReleaseMemObject(d_all_outputs_);
    clReleaseMemObject(d_prev_deltas_);
    clReleaseMemObject(d_deltas_);
    clReleaseMemObject(d_delta_weights_);
    clReleaseMemObject(d_output_);
    clReleaseMemObject(d_temp_output_);

    // Release kernels.
    clReleaseKernel(layer_propagate_forward_kernel_);
    clReleaseKernel(layer_propagate_backwards_kernel_);
    clReleaseKernel(apply_delta_weights_kernel_);
}

void NN_BP_OpenCL::printInfo()
{
    NN::printInfo();
    NN_OpenCL::printInfo();
    bpParameters_.PrintInfo();

}

// Initializes random weights on host and copies it to device.
void NN_BP_OpenCL::initializeWeights()
{
    CL_FINISH
    randomizeVector(h_weights_,bpParameters_.lowerLimit, bpParameters_.upperLimit);
    // write weights to device
    int err = clEnqueueWriteBuffer(queue_, d_weights_, CL_FALSE, 0, h_weights_.size()*sizeof(float), &h_weights_[0], 0, NULL, NULL);
    if(err != CL_SUCCESS) error("Error writing weights to buffer",151);
    CL_FINISH
}

// Set delta weights on device to 0.
void NN_BP_OpenCL::initializeDeltaWeights()
{
    //for(int i=0;i<getNumWeights();i++)
    //    h_delta_weights_[i] = 0;

    // write delta weights to device
    //int err = clEnqueueWriteBuffer(queue_, d_delta_weights_, CL_FALSE, 0, getNumWeights()*sizeof(float), h_delta_weights_, 0, NULL, NULL);
    //if(err != CL_SUCCESS) error("Error writing delta weights to buffer",155);

}

// Propagates input samples (the one with index sampleIndex) and sets all the outputs from all the neurons.
void NN_BP_OpenCL::propagateForward(vector<float> &inputSamples, int sampleIndex)
{
    CL_FINISH

    int err;

    cl_mem d_inputSamples = vectorToClmem(inputSamples);
    size_t sampleOffset = sampleIndex * getInputSize() * sizeof(float); // in bytes

    cl_int neuron_index = 0;

    // output from first neurons == input
    err = clEnqueueCopyBuffer(queue_, d_inputSamples, d_all_outputs_, sampleOffset, 0, getInputSize() * sizeof(float), 0, NULL, NULL);
    if(err != CL_SUCCESS) error("Error writing delta weights to buffer",155);

    neuron_index += h_layers_[0];

    cl_mem d_input = d_inputSamples;
    cl_mem d_output;
    cl_int input_offset = 0;
    cl_int weights_offset = 0;

    for(int i=1;i<getNumLayers();i++)  // i==1 => input -> xxx
    {
        input_offset=0;
        if((getNumLayers()-i)%2==1)
        {
            d_input = d_temp_output_;
            d_output = d_output_;

        }
        else
        {
            d_input = d_output_;
            d_output = d_temp_output_;
        }
        if(i==1)
        {
            d_input = d_inputSamples;
            input_offset = sampleOffset/sizeof(float); // in elements
        }

        // set kernel arguments
        err = clSetKernelArg(layer_propagate_forward_kernel_, 0, sizeof(cl_mem), &d_input);
        err|= clSetKernelArg(layer_propagate_forward_kernel_, 1, sizeof(cl_int), &input_offset);
        err|= clSetKernelArg(layer_propagate_forward_kernel_, 3, sizeof(cl_int), &weights_offset);
        err|= clSetKernelArg(layer_propagate_forward_kernel_, 4, sizeof(cl_mem), &d_output);
        err|= clSetKernelArg(layer_propagate_forward_kernel_, 6, sizeof(cl_int), &neuron_index);
        err|= clSetKernelArg(layer_propagate_forward_kernel_, 7, sizeof(int), &h_layers_[i-1]);
        err|= clSetKernelArg(layer_propagate_forward_kernel_, 8, sizeof(int), &h_layers_[i]);
        cl_int isLastLayer = (i == getNumLayers()-1) ? 1 : 0;
        err|= clSetKernelArg(layer_propagate_forward_kernel_, 9, sizeof(cl_int), &isLastLayer);
        if(err != CL_SUCCESS) error("Error setting kernel arguments for propagate layer forward kernel in propagate forward.", 557);

        // run kernel
        size_t localWorkSize = 64;
        size_t globalWorkSize = ( (h_layers_[i] + localWorkSize - 1) / localWorkSize ) * localWorkSize;
        err = clEnqueueNDRangeKernel(queue_, layer_propagate_forward_kernel_, 1, NULL, &globalWorkSize,&localWorkSize, 0, NULL,NULL);
        if(err != CL_SUCCESS) error("Error running kernel propagate layer forward.", 357);

        // adjust offsets
        neuron_index += h_layers_[i];
        weights_offset += h_layers_[i]*(h_layers_[i-1]+1);
        sampleOffset = 0; // this was just for input from samples
    }

    CL_FINISH
}


// propagates error back through the network and sets the delta weights
void NN_BP_OpenCL::propagateBackwards(vector<float> &outputSamples, int sampleIndex)
{
    CL_FINISH
    int err;

    //cl_mem d_targetSamples = vectorToClmem(outputSamples);
    if(outputSamples != h_training_targets_) error("propagateBackwards argument must be training_targets");


    cl_int targets_offset = sampleIndex * getOutputSize();

    cl_int all_outputs_offset = getNumNeurons();
    cl_int weights_offset = getNumWeights();
    cl_int isLastLayer = 1;

    cl_mem prev_deltas;
    cl_mem deltas;
    for(int layer=getNumLayers()-1; layer>0; layer--)
    {
        all_outputs_offset -= h_layers_[layer];
        err = clSetKernelArg(layer_propagate_backwards_kernel_, 1, sizeof(cl_int), &all_outputs_offset);

        // doesn't matter which one is the first and where the last deltas will end up, they just have to alternate
        if(layer%2==0)
        {
            prev_deltas = d_prev_deltas_;
            deltas = d_deltas_;
        }
        else
        {
            prev_deltas = d_deltas_;
            deltas = d_prev_deltas_;
        }
        err|= clSetKernelArg(layer_propagate_backwards_kernel_, 2, sizeof(cl_mem), &prev_deltas);
        err|= clSetKernelArg(layer_propagate_backwards_kernel_, 3, sizeof(cl_mem), &deltas);

        weights_offset -= h_layers_[layer] * (h_layers_[layer-1] + 1);
        err |= clSetKernelArg(layer_propagate_backwards_kernel_, 5, sizeof(cl_int), &weights_offset);
        err |= clSetKernelArg(layer_propagate_backwards_kernel_, 7, sizeof(cl_int), &h_layers_[layer-1]);
        err |= clSetKernelArg(layer_propagate_backwards_kernel_, 8, sizeof(cl_int), &h_layers_[layer]);
        if(isLastLayer) err |= clSetKernelArg(layer_propagate_backwards_kernel_, 9, sizeof(cl_int), &h_layers_[layer]);
        else err |= clSetKernelArg(layer_propagate_backwards_kernel_, 9, sizeof(cl_int), &h_layers_[layer+1]);
        err |= clSetKernelArg(layer_propagate_backwards_kernel_, 11, sizeof(cl_int), &isLastLayer);
        isLastLayer = 0;// just the first iteration
        err |= clSetKernelArg(layer_propagate_backwards_kernel_, 13, sizeof(cl_int), &targets_offset);

        if(err != CL_SUCCESS) error("Error setting kernel arguments for propagate layer backwards kernel in propagate forward.", 551);

        // run kernel
        size_t localWorkSize = 64;
        size_t globalWorkSize = ( (h_layers_[layer] + localWorkSize - 1) / localWorkSize ) * localWorkSize;

        err = clEnqueueNDRangeKernel(queue_, layer_propagate_backwards_kernel_, 1, NULL, &globalWorkSize,&localWorkSize, 0, NULL,NULL);
        if(err != CL_SUCCESS) error("Error running kernel propagate layer backwards.", 357);

    }

    // delta weights are now set

    CL_FINISH

}

// Adds the weights on device and copies them back to host.
void NN_BP_OpenCL::applyDeltaWeights()
{
    CL_FINISH
    int err;
    // run kernel
    size_t localWorkSize = 256;
    size_t globalWorkSize = ( ( getNumWeights() + localWorkSize - 1) / localWorkSize ) * localWorkSize;
    err = clEnqueueNDRangeKernel(queue_, apply_delta_weights_kernel_, 1, NULL, &globalWorkSize,&localWorkSize, 0, NULL,NULL);
    if(err != CL_SUCCESS) error("Error running apply_delta_weights_kernel_.", 357);

    // copy back weights
    err = clEnqueueReadBuffer(queue_, d_weights_, CL_TRUE, 0, getNumWeights()*sizeof(float), &h_weights_[0], 0, NULL, NULL);
    if(err != CL_SUCCESS) error("Error reading back weights", 1671);
    CL_FINISH
}

void NN_BP_OpenCL::simulateNeuralNetwork(vector<float> &inputSamples, int sampleIndex)
{
    int err;
    propagateForward(inputSamples, sampleIndex);
    // output is now in d_ouput_

    // copy it back to h_outputs_
    err = clEnqueueReadBuffer(queue_, d_output_, CL_TRUE, 0, getOutputSize()*sizeof(float), &h_outputs_[0], 0, NULL, NULL);
    if(err != CL_SUCCESS) error("Error reading back weights", 1671);
}

NN_BP_OpenCL_GPU::NN_BP_OpenCL_GPU(string nnetConfigFile, string trainingDataFile, string bpConfigFile) :
    NN(nnetConfigFile,trainingDataFile), NN_BP_OpenCL(nnetConfigFile,trainingDataFile,bpConfigFile, CL_DEVICE_TYPE_GPU)
{
}

NN_BP_OpenCL_CPU::NN_BP_OpenCL_CPU(string nnetConfigFile, string trainingDataFile, string bpConfigFile) :
    NN(nnetConfigFile,trainingDataFile), NN_BP_OpenCL(nnetConfigFile,trainingDataFile,bpConfigFile, CL_DEVICE_TYPE_CPU)
{
}


// NN_reloaded

// Loads a file with weights
NN_reloaded::NN_reloaded(std::string weightsFile)
{
    ifstream file(weightsFile.c_str(), ios::in | ios::binary);
    if(!file.is_open()) error(string("Error opening file: ") + weightsFile);

    // read numLayers
    int numLayers;
    file.read((char *)&numLayers, sizeof(int));

    // read layers
    h_layers_.resize(numLayers);
    file.read((char *)&h_layers_[0], numLayers * sizeof(int));

    // read weights
    h_weights_.resize(getNumWeights());
    file.read((char *)&h_weights_[0], getNumWeights() * sizeof(float));

    // read input and output ranges
    in_ranges_.resize(2 * getInputSize());
    h_out_ranges_.resize(2 * getOutputSize());
    file.read((char *)&in_ranges_[0], 2*getInputSize() * sizeof(float));
    file.read((char *)&h_out_ranges_[0], 2*getOutputSize() * sizeof(float));

    file.close();

    // allocate memory for CPU simulation
    h_output_ = new float[getWidestLayer()];
    h_temp_output_ = new float[getWidestLayer()];
}

double NN_reloaded::train()
{
    error("Training not allowed");
    return 0;
}

double NN_reloaded::timing(int numIterations)
{
    return 0; // didn't feel like it
}
void NN_reloaded::printInfo()
{
    cout << "Neural network info" << endl << "-------------------" << endl;
    cout << "Layers: ";
    for(int i=0;i<getNumLayers();i++) cout << h_layers_[i] << " ";
    cout << "\nInput ranges:\n";
    for(int i=0;i<getInputSize();i++) cout << "[" << in_ranges_[2*i] << "," << in_ranges_[2*i+1] << "]" << endl;
    cout << "Output ranges:\n";
    for(int i=0;i<getOutputSize();i++) cout << "[" << h_out_ranges_[2*i] << "," << h_out_ranges_[2*i+1] << "]" << endl;
    cout << "Total weights: " << getNumWeights() << endl;
}
